@@ -3,7 +3,7 @@ package YAHC;
 use strict;
 use warnings;
 
-our $VERSION = '0.006';
+our $VERSION = '0.007';
 
 use EV;
 use Time::HiRes;
@@ -47,11 +47,13 @@ our @EXPORT_OK = qw/
     yahc_conn_last_error
     yahc_conn_id
     yahc_conn_url
+    yahc_conn_target
     yahc_conn_state
     yahc_conn_errors
     yahc_conn_timeline
     yahc_conn_request
     yahc_conn_response
+    yahc_conn_retries_left
 /;
 
 our %EXPORT_TAGS = (all => \@EXPORT_OK);
@@ -112,7 +114,7 @@ sub request {
         ($conn_id, $request) = @args;
     }
 
-    die "YAHC: Connection with name '$conn_id' already exists"
+    die "YAHC: Connection with name '$conn_id' already exists\n"
         if exists $self->{connections}{$conn_id};
 
     my $conn = {
@@ -130,10 +132,10 @@ sub request {
     my $pool_args = $self->{pool_args};
     $request->{_target} = _wrap_target_selection($request->{host}) if $request->{host};
     do { $request->{$_} ||= $pool_args->{$_} if $pool_args->{$_} } foreach (qw/host port scheme/);
-    die 'YAHC: host must be defined' unless $request->{host};
+    die "YAHC: host must be defined\n" unless $request->{host};
 
     $request->{scheme} //= 'http';
-    die 'YAHC: only support scheme http' unless $request->{scheme} eq 'http';
+    die "YAHC: only support scheme http\n" unless $request->{scheme} eq 'http';
 
     my %callbacks;
     foreach (@{ CALLBACKS() }) {
@@ -151,9 +153,22 @@ sub request {
     return $conn;
 }
 
-sub run      { $_[0]->_run(0, $_[1])       }
-sub run_once { $_[0]->_run(EV::RUN_ONCE)   }
-sub run_tick { $_[0]->_run(EV::RUN_NOWAIT) }
+sub drop {
+    my ($self, $c) = @_;
+    my $conn_id = ref($c) eq 'HASH' ? $c->{id} : $c;
+
+    my $conn = $self->{connections}{$conn_id}
+      or return;
+
+    _register_in_timeline($conn, "dropping connection from pool") if $conn->{keep_timeline};
+    $self->_set_completed_state($conn_id) unless $conn->{state} == YAHC::State::COMPLETED();
+
+    return delete $self->{connections}{$conn_id};
+}
+
+sub run      { shift->_run(0, @_)          }
+sub run_once { shift->_run(EV::RUN_ONCE)   }
+sub run_tick { shift->_run(EV::RUN_NOWAIT) }
 
 ################################################################################
 # Routines to manipulate connections (also user facing)
@@ -161,6 +176,9 @@ sub run_tick { $_[0]->_run(EV::RUN_NOWAIT) }
 
 sub yahc_reinit_conn {
     my ($conn, $args) = @_;
+    die "YAHC: cannot reinit completed connection\n"
+        if $conn->{state} >= YAHC::State::COMPLETED();
+
     $conn->{attempt} = 0;
     $conn->{state} = YAHC::State::INITIALIZED();
     return unless defined $args && ref($args) eq 'HASH';
@@ -182,11 +200,19 @@ sub yahc_conn_errors        { $_[0]->{errors}   }
 sub yahc_conn_timeline      { $_[0]->{timeline} }
 sub yahc_conn_request       { $_[0]->{request}  }
 sub yahc_conn_response      { $_[0]->{response} }
+sub yahc_conn_retries_left  { $_[0]->{retries} - $_[0]->{attempt} }
+
+sub yahc_conn_target {
+    my $target = $_[0]->{selected_target};
+    return unless $target && scalar @{ $target };
+    my ($host, $ip, $port) = @{ $target };
+    return ($host || $ip) . ($port ne "80" ? ":$port" : '');
+}
 
 sub yahc_conn_url {
     my $target = $_[0]->{selected_target};
     my $request = $_[0]->{request};
-    return unless $target;
+    return unless $target && @{ $target };
 
     my ($host, $ip, $port) = @{ $target };
     return $request->{scheme} . "://"
@@ -201,8 +227,8 @@ sub yahc_conn_url {
 ################################################################################
 
 sub _run {
-    my ($self, $how, $until_state, $until_state_connection_ids) = @_;
-    die 'YAHC: storage object is destroyed' unless $self->{storage};
+    my ($self, $how, $until_state, @cs) = @_;
+    die "YAHC: storage object is destroyed\n" unless $self->{storage};
 
     if ($self->{pid} != $$) {
         _log_message('Reinitializing event loop after forking') if $self->{debug};
@@ -212,19 +238,16 @@ sub _run {
 
     if (defined $until_state) {
         my $until_state_str = _strstate($until_state);
-        die "YAHC: unknown until_state $until_state" if $until_state_str =~ m/unknown/;
+        die "YAHC: unknown until_state $until_state\n" if $until_state_str =~ m/unknown/;
 
-        my @connections;
-        if (defined $until_state_connection_ids) {
-            @connections = @{ $until_state_connection_ids };
-            @connections = grep { exists $self->{connections}{$_} ? $self->{connections}{$_} : () } @connections;
-        } else {
-            @connections = values %{ $self->{connections} }; # all connections
-        }
+        my @connections = (@cs == 0)
+                        ? values %{ $self->{connections} }
+                        : map { $self->{connections}{$_} || () }
+                          map { ref($_) eq 'HASH' ? $_->{id} : $_ } @cs;
 
         $self->{stop_condition} = {
             expected_state => $until_state,
-            connections => { map { $_->id => 1 } grep { $_->state < $until_state } @connections },
+            connections => { map { $_->{id} => 1 } grep { $_->{state} < $until_state } @connections },
         };
     } else {
         delete $self->{stop_condition};
@@ -251,7 +274,7 @@ sub _check_stop_condition {
     my $stop_condition = $self->{stop_condition};
     return if !$stop_condition || $conn->{state} < $stop_condition->{expected_state};
 
-    delete $stop_condition->{connections}{$conn->id};
+    delete $stop_condition->{connections}{$conn->{id}};
     my $awaiting_connections = scalar keys %{ $stop_condition->{connections} };
     my $expected_state = $stop_condition->{expected_state};
 
@@ -272,7 +295,7 @@ sub _set_init_state {
 
     my $conn = $self->{connections}{$conn_id};
     my $watchers = $self->{watchers}{$conn_id};
-    die "YAHC: unknown connection id $conn_id" unless $conn && $watchers;
+    die "YAHC: unknown connection id $conn_id\n" unless $conn && $watchers;
 
     $conn->{response} = { status_code => 0 };
     $conn->{state} = YAHC::State::INITIALIZED();
@@ -315,7 +338,7 @@ sub _set_wait_synack_state {
 
     my $conn = $self->{connections}{$conn_id};
     my $watchers = $self->{watchers}{$conn_id};
-    die "YAHC: unknown connection id $conn_id" unless $conn && $watchers;
+    die "YAHC: unknown connection id $conn_id\n" unless $conn && $watchers;
     _assert_state($conn, YAHC::State::INITIALIZED()) if $conn->{debug};
 
     $conn->{state} = YAHC::State::WAIT_SYNACK();
@@ -353,7 +376,7 @@ sub _set_write_state {
 
     my $conn = $self->{connections}{$conn_id};
     my $watcher = $self->{watchers}{$conn_id}{io};
-    die "YAHC: unknown connection id $conn_id" unless $conn && $watcher;
+    die "YAHC: unknown connection id $conn_id\n" unless $conn && $watcher;
     _assert_connected($conn) if $conn->{debug};
 
     $conn->{state} = YAHC::State::WRITING();
@@ -369,7 +392,7 @@ sub _set_write_state {
     my $write_cb = $self->_get_safe_wrapper($conn, sub {
         my $wlen = POSIX::write($fd, $buf, $length);
 
-        if (!$wlen) {
+        if (!defined $wlen || $wlen == 0) {
             return if $! == EWOULDBLOCK || $! == EAGAIN || $! == EINTR;
 
             my $error = !defined $wlen
@@ -395,7 +418,7 @@ sub _set_read_state {
 
     my $conn = $self->{connections}{$conn_id};
     my $watcher = $self->{watchers}{$conn_id}{io};
-    die "YAHC: unknown connection id $conn_id" unless $conn && $watcher;
+    die "YAHC: unknown connection id $conn_id\n" unless $conn && $watcher;
     _assert_connected($conn) if $conn->{debug};
 
     $conn->{state} = YAHC::State::READING();
@@ -411,14 +434,14 @@ sub _set_read_state {
     my $read_cb = $self->_get_safe_wrapper($conn, sub {
         my $rlen = POSIX::read($fd, my $b = '', TCP_READ_CHUNK);
 
-        if (!$rlen) {
+        if (!defined $rlen || $rlen == 0) {
             return if $! == EWOULDBLOCK || $! == EAGAIN || $! == EINTR;
             if (not defined $rlen) {
                 _register_error($conn, YAHC::Error::READ_ERROR(), "Failed to receive TCP data: $!");
             } elsif ($content_length > 0) { # i.e. rlen == 0 and $content_length > 0
-                register_error($conn, YAHC::Error::READ_ERROR(), "Premature EOF, expect %d bytes more", $content_length - length($buf));
+                _register_error($conn, YAHC::Error::READ_ERROR(), "Premature EOF, expect %d bytes more", $content_length - length($buf));
             } else { # i.e. rlen == 0
-                register_error($conn, YAHC::Error::READ_ERROR(), "Premature EOF");
+                _register_error($conn, YAHC::Error::READ_ERROR(), "Premature EOF");
             }
 
             $self->_set_init_state($conn_id);
@@ -427,7 +450,7 @@ sub _set_read_state {
             if (!$decapitated && ($neck_pos = index($buf, "${CRLF}${CRLF}")) > 0) {
                 my $headers = _parse_http_headers($conn, substr($buf, 0, $neck_pos, ''));
                 if (not defined $headers) {
-                    $self->_set_user_action_state($conn_id, YAHC::Error::RESPONSE_ERROR(), "unsuported HTTP reponse");
+                    $self->_set_user_action_state($conn_id, YAHC::Error::RESPONSE_ERROR(), "unsupported HTTP reponse");
                     return;
                 }
 
@@ -533,27 +556,27 @@ sub _build_socket_and_connect {
     my ($ip, $port, $timeouts) = @_;
 
     my $sock;
-    socket($sock, PF_INET, SOCK_STREAM, 0) or die "YAHC: Failed to construct TCP socket: $!";
+    socket($sock, PF_INET, SOCK_STREAM, 0) or die "YAHC: Failed to construct TCP socket: $!\n";
 
-    my $flags = fcntl($sock, F_GETFL, 0) or die "YAHC: Failed to get fcntl F_GETFL flag: $!";
-    fcntl($sock, F_SETFL, $flags | O_NONBLOCK) or die "YAHC: Failed to set fcntl O_NONBLOCK flag: $!";
+    my $flags = fcntl($sock, F_GETFL, 0) or die "YAHC: Failed to get fcntl F_GETFL flag: $!\n";
+    fcntl($sock, F_SETFL, $flags | O_NONBLOCK) or die "YAHC: Failed to set fcntl O_NONBLOCK flag: $!\n";
 
     if (my $read_timeout = $timeouts->{read_timeout}) {
         my $read_struct  = pack('l!l!', $read_timeout, 0);
         setsockopt($sock, SOL_SOCKET, SO_RCVTIMEO, $read_struct)
-          or die "YAHC: Failed to set setsockopt(SO_RCVTIMEO): $!";
+          or die "YAHC: Failed to set setsockopt(SO_RCVTIMEO): $!\n";
     }
 
     if (my $write_timeout = $timeouts->{write_timeout}) {
         my $write_struct = pack('l!l!', $write_timeout, 0);
         setsockopt($sock, SOL_SOCKET, SO_SNDTIMEO, $write_struct )
-          or die "YAHC: Failed to set setsockopt(SO_SNDTIMEO): $!";
+          or die "YAHC: Failed to set setsockopt(SO_SNDTIMEO): $!\n";
     }
 
     my $ip_addr = inet_aton($ip);
     my $addr = pack_sockaddr_in($port, $ip_addr);
     if (!connect($sock, $addr) && $! != EINPROGRESS) {
-        die "YAHC: Failed to connect to $ip: $!";
+        die "YAHC: Failed to connect to $ip: $!\n";
     }
 
     return $sock;
@@ -566,7 +589,7 @@ sub _get_next_target {
     # TODO STATE_RESOLVE_DNS
     ($host, $port) = ($1, $2) if $host =~ m/^(.+):([0-9]+)$/o;
     $ip = $host if $host =~ m/^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$/o;
-    $ip = inet_ntoa(gethostbyname($host) or die "Failed to resolve $host: $!") unless $ip;
+    $ip = inet_ntoa(gethostbyname($host) or die "Failed to resolve '$host': $!\n") unless $ip;
     $port ||= $conn->{request}{port} || HTTP_PORT;
 
     return @{ $conn->{selected_target} = [ $host, $ip, $port ] };
@@ -581,7 +604,7 @@ sub _set_request_timer {
 
     my $conn = $self->{connections}{$conn_id};
     my $watchers = $self->{watchers}{$conn_id};
-    die "YAHC: unknown connection id $conn_id" unless $conn && $watchers;
+    die "YAHC: unknown connection id $conn_id\n" unless $conn && $watchers;
 
     if (my $timer = delete $watchers->{connection_timer}) {
         $timer->stop;
@@ -609,7 +632,7 @@ sub _set_connection_timer {
 
     my $conn = $self->{connections}{$conn_id};
     my $watchers = $self->{watchers}{$conn_id};
-    die "YAHC: unknown connection id $conn_id" unless $conn && $watchers;
+    die "YAHC: unknown connection id $conn_id\n" unless $conn && $watchers;
 
     if (my $timer = delete $watchers->{connection_timer}) {
         $timer->stop;
@@ -691,7 +714,7 @@ sub _wrap_target_selection {
 
     my $idx = 0;
     return sub { $host->[$idx++ % @$host]; } if $ref eq 'ARRAY';
-    die "YAHC: unsupported host format";
+    die "YAHC: unsupported host format\n";
 }
 
 sub _call_state_callback {
@@ -736,20 +759,20 @@ sub _register_error {
     my $strerror = sprintf("$format", @arguments);
     $strerror =~ s/\s+$//g;
     _register_in_timeline($conn, "error=$error ($strerror)") if $conn->{debug};
-    push @{ $conn->{errors} ||= [] }, [ $error, $strerror, Time::HiRes::time ];
+    push @{ $conn->{errors} ||= [] }, [ $error, $strerror, yahc_conn_target($conn) // 'no_target_yet', Time::HiRes::time ];
 }
 
 sub _assert_state {
     my ($conn, $expected_state) = @_;
     return $conn->{state} == $expected_state;
-    die sprintf("YAHC connection '%s' is in unexpected state %s, expected %s",
+    die sprintf("YAHC connection '%s' is in unexpected state %s, expected %s\n",
                 $conn->{id}, _strstate($conn->{state}), _strstate($expected_state));
 }
 
 sub _assert_connected {
     my $conn = shift;
     return if $conn->{state} >= YAHC::State::CONNECTED() && $conn->{state} < YAHC::State::COMPLETED();
-    die sprintf("YAHC connection '%s' expected to be in connected state, but it's in %s",
+    die sprintf("YAHC connection '%s' expected to be in connected state, but it's in %s\n",
                 $conn->{id}, _strstate($conn->{state}));
 }
 
